/**
 * processor.c
 *
 * Copyright (C) 2003 - 2008 Bojan Smojver, Rexursive
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "private.h"
#include "scanner.h"

apr_status_t rxv_spin_build(rxv_spin_extra_t *extra){
  /* nothing to do */
  if(APR_BRIGADE_EMPTY(extra->input))
    return APR_SUCCESS;

  /* we have to set up scanner */
  if(!extra->scanner){
    if(rxv_spin_yylex_init(&extra->scanner))
      return apr_get_os_error();

    rxv_spin_yyset_extra(extra,extra->scanner);

    extra->scan=APR_BRIGADE_FIRST(extra->input);

    /* make yywrap() read the buckets */
    memset(extra->buf,0,2);
    rxv_spin_yy_scan_buffer(extra->buf,2,extra->scanner);
  }

  if(rxv_spin_yyparse(extra->scanner)){
    rxv_spin_yylex_destroy(extra->scanner);
    extra->scanner=NULL;
    return APR_EGENERAL;
  }

  /* bumped into EOF, maybe need to back up */
  if(extra->eof){
    if(extra->offset>extra->pos){
      apr_bucket *b;

      /* not at the beginning - split */
      if(extra->pos>0){
        if(apr_brigade_partition(extra->input,extra->pos,&b)!=APR_SUCCESS){
          extra->error="cannot partition brigade";

          rxv_spin_yylex_destroy(extra->scanner);
          extra->scanner=NULL;

          return APR_EGENERAL;
        }
        extra->split=apr_brigade_split(extra->input,b);
      } else{
        extra->split=extra->input;
      }
      extra->input=NULL;
    } else if(extra->bpos>=extra->blen){
      extra->input=NULL;
    }
  }

  /* bumped into EOS, see if we're done */
  if(extra->eos){
    if(extra->offset>extra->pos){
      rxv_spin_yylex_destroy(extra->scanner);
      extra->scanner=NULL;

      return APR_EGENERAL;
    } else if(extra->bpos>=extra->blen){
      extra->input=NULL;
    }
  }

  if(!extra->input){
    rxv_spin_yylex_destroy(extra->scanner);
    extra->scanner=NULL;
  }

  return APR_SUCCESS;
}

static rxv_spin_data_t *ref_resolve(rxv_spin_ref_t *r,rxv_spin_ctx_t *ctx){
  rxv_spin_data_t *data,*prev;

  if(r->ref){
    prev=r->ref->data;

    data=ref_resolve(r->ref,ctx); /* recursively resolve to the top first */

    if(data){
      if(!r->data || data!=prev){
        if(data->size==RXV_SPIN_ROWS){
          apr_array_header_t *a;

          a=apr_hash_get(data->cols,r->name,APR_HASH_KEY_STRING);
          r->data=apr_is_empty_array(a)?NULL:(rxv_spin_data_t *)a->elts;
        } else{
          r->data=NULL;
        }
      }
    } else{
      r->data=NULL;
    }
  } else if(!r->data){
    r->data=apr_hash_get(ctx->data->cols,r->name,APR_HASH_KEY_STRING);
  }

  if(r->data){
    if(r->ip)
      return r->data+*r->ip;
    else
      return r->data;
  }

   return NULL;
}

apr_status_t rxv_spin_render(rxv_spin_node_t *node,rxv_spin_extra_t *extra){
  char *buf;
  apr_size_t val;
  unsigned long num;
  apr_status_t rv;
  apr_bucket_brigade *bb=extra->output;
  apr_bucket *b;
  rxv_spin_txt_t *txt;
  rxv_spin_ref_t *ref,*rfr;
  rxv_spin_for_t *loop;
  rxv_spin_if_t *cond;
  apr_pool_t *p=extra->pool;
  rxv_spin_data_t *data,*rdta;

  for(;node;node=node->next){
    switch(node->type){
      case RXV_SPIN_TXT:
        txt=(rxv_spin_txt_t *)node;

        /* we already have this in a bucket, copy */
        if(txt->used){
          apr_bucket_copy(txt->bucket,&b);
        } else{
          APR_BUCKET_REMOVE(txt->bucket);
          b=txt->bucket;
          txt->used=RXV_SPIN_TRUE;
        }

        APR_BRIGADE_INSERT_TAIL(bb,b);

        break;
      case RXV_SPIN_REF:
        ref=(rxv_spin_ref_t *)node;

        data=ref_resolve(ref,extra->ctx);

        /* skip rows, we cannot insert them into text */
        if(data){
          char *buf=NULL;
          apr_size_t len=0;

          /* prepare data and size for bucket */
          switch(ref->type){
            case RXV_SPIN_REF_REG:
              if(data->size==RXV_SPIN_ROWS)
                continue;
              buf=data->data;
              len=data->size;
              break;
            case RXV_SPIN_REF_SIZ:
              buf=apr_psprintf(p,"%" APR_SIZE_T_FMT,rxv_spin_size(data));
              len=strlen(buf);
              break;
            case RXV_SPIN_REF_IND:
              if(ref->ip)
                buf=apr_psprintf(p,"%" APR_SIZE_T_FMT,*ref->ip+1);
              else
                buf="0";
              len=strlen(buf);
              break;
            default: /* must be an error, we don't know this */
              return APR_EGENERAL;
          }

          /* only replace valid stuff (i.e. size, index and singles) */
          if(buf && len>0){
            b=apr_bucket_heap_create(buf,len,NULL,extra->output->bucket_alloc);
            APR_BRIGADE_INSERT_TAIL(bb,b);
          }
        }
        break;
      case RXV_SPIN_FOR:
        loop=(rxv_spin_for_t *)node;
        ref=loop->ref;

        data=ref_resolve(ref,extra->ctx);

        if(data){
          if(data->size==RXV_SPIN_ROWS){
            if(data->cols){
              apr_size_t size=rxv_spin_size(data);
            
              for(ref->index=0;ref->index<size;ref->index++)
                if((rv=rxv_spin_render(loop->stmt,extra))!=APR_SUCCESS)
                  return rv;
            }
          } else{
            if(data->data)
              if((rv=rxv_spin_render(loop->stmt,extra))!=APR_SUCCESS)
                return rv;
          }
        }
        break;
      case RXV_SPIN_IF:
        cond=(rxv_spin_if_t *)node;
        ref=cond->ref;
        rfr=cond->rfr;

        data=ref_resolve(ref,extra->ctx);

        if(data){
          unsigned int truth;
 
          switch(cond->oper){
            case RXV_SPIN_OP_NONE:
              switch(ref->type){
                case RXV_SPIN_REF_REG:
                  truth=(data->size==RXV_SPIN_ROWS && data->cols) || data->data;
                  break;
                case RXV_SPIN_REF_SIZ:
                  truth=(rxv_spin_size(data)>0); break;
                case RXV_SPIN_REF_IND:
                  truth=(ref->ref!=NULL); break;
                default: /* don't know this, false */
                  truth=RXV_SPIN_FALSE; break;
              }
              break;
            case RXV_SPIN_OP_MATCH:
              switch(ref->type){
                case RXV_SPIN_REF_REG:
                  truth=(data->size!=RXV_SPIN_ROWS && data->data &&
                         !ap_regexec(cond->reg,data->data,0,NULL,0)); break;
                case RXV_SPIN_REF_SIZ:
                  buf=apr_psprintf(p,"%" APR_SIZE_T_FMT,rxv_spin_size(data));
                  truth=!ap_regexec(cond->reg,buf,0,NULL,0);
                  break;
                case RXV_SPIN_REF_IND:
                  if(ref->ip)
                    buf=apr_psprintf(p,"%" APR_SIZE_T_FMT,*ref->ip+1);
                  else
                    buf="0";
                  truth=!ap_regexec(cond->reg,buf,0,NULL,0);
                  break;
                default: /* must be an error, we don't know this */
                  return APR_EGENERAL;
              }
              break;
            case RXV_SPIN_OP_EQ_LIT:
              switch(ref->type){
                case RXV_SPIN_REF_REG:
                  truth=(data->size!=RXV_SPIN_ROWS && data->data &&
                         !strcmp(data->data,cond->lit));
                  break;
                case RXV_SPIN_REF_SIZ:
                  buf=apr_psprintf(p,"%" APR_SIZE_T_FMT,rxv_spin_size(data));
                  truth=!strcmp(buf,cond->lit);
                  break;
                case RXV_SPIN_REF_IND:
                  if(ref->ip)
                    buf=apr_psprintf(p,"%" APR_SIZE_T_FMT,*ref->ip+1);
                  else
                    buf="0";
                  truth=!strcmp(buf,cond->lit);
                  break;
                default: /* must be an error, we don't know this */
                  return APR_EGENERAL;
              }
              break;
            case RXV_SPIN_OP_EQ_NUM:
              switch(ref->type){
                case RXV_SPIN_REF_REG:
                  if(data->size!=RXV_SPIN_ROWS && data->data){
                    num=0; sscanf(data->data,"%lu",&num);
                    truth=(num==cond->num);
                  } else{
                    truth=RXV_SPIN_FALSE;
                  }
                  break;
                case RXV_SPIN_REF_SIZ:
                  truth=(rxv_spin_size(data)==cond->num); break;
                case RXV_SPIN_REF_IND:
                  if(ref->ip)
                    truth=(*ref->ip+1==cond->num);
                  else
                    truth=(0==cond->num);
                  break;
                default: /* must be an error, we don't know this */
                  return APR_EGENERAL;
              }
              break;
            case RXV_SPIN_OP_EQ_REF:
              if(!rfr){ /* impossible reference, always false */
                if(cond->neg &&
                   (rv=rxv_spin_render(cond->neg,extra))!=APR_SUCCESS)
                  return rv;

                continue;
              }

              rdta=ref_resolve(rfr,extra->ctx);

              if(!rdta){ /* no data, always false */
                if(cond->neg &&
                   (rv=rxv_spin_render(cond->neg,extra))!=APR_SUCCESS)
                  return rv;

                continue;
              }

              /* we are wasting CPU cycles and memory here as a tradeoff
                 for slightly simpler code - life isn't perfect ;-) */
              switch(rfr->type){
                case RXV_SPIN_REF_REG:
                  if(rdta->size==RXV_SPIN_ROWS){
                    buf=NULL; val=0;
                  } else{
                    buf=rdta->data;
                    if(buf){
                      val=0; sscanf(buf,"%" APR_SIZE_T_FMT,&val);
                    } else{
                      val=0;
                    }
                  }
                  break;
                case RXV_SPIN_REF_SIZ:
                  val=rxv_spin_size(rdta);
                  buf=apr_psprintf(p,"%" APR_SIZE_T_FMT,val);
                  break;
                case RXV_SPIN_REF_IND:
                  if(rfr->ip){
                    val=*rfr->ip+1;
                    buf=apr_psprintf(p,"%" APR_SIZE_T_FMT,val);
                  } else{
                    val=0; buf="0";
                  }
                  break;
                default: /* must be an error, we don't know this */
                  return APR_EGENERAL;
              }

              switch(ref->type){
                case RXV_SPIN_REF_REG:
                  truth=(data->size!=RXV_SPIN_ROWS && data->data && buf &&
                         !strcmp(data->data,buf));
                  break;
                case RXV_SPIN_REF_SIZ:
                  truth=(rxv_spin_size(data)==val); break;
                case RXV_SPIN_REF_IND:
                  if(ref->ip)
                    truth=(*ref->ip+1==val);
                  else
                    truth=(0==val);
                  break;
                default: /* must be an error, we don't know this */
                  return APR_EGENERAL;
              }
              break;
            case RXV_SPIN_OP_MODULO:
              switch(ref->type){
                case RXV_SPIN_REF_REG:
                  if(data->size!=RXV_SPIN_ROWS && data->data){
                    num=0; sscanf(data->data,"%lu",&num);
                    truth=num%cond->num==cond->res;
                  } else{
                    truth=RXV_SPIN_FALSE;
                  }
                  break;
                case RXV_SPIN_REF_SIZ:
                  truth=(rxv_spin_size(data)%cond->num==cond->res); break;
                case RXV_SPIN_REF_IND:
                  if(ref->ip)
                    truth=(*ref->ip+1%cond->num==cond->res);
                  else
                    truth=(0%cond->num==cond->res);
                  break;
                default: /* must be an error, we don't know this */
                  return APR_EGENERAL;
              }
              break;
            default: /* must be an error, we don't know this */
              return APR_EGENERAL;
          }

          if(truth){
            if(cond->pos && (rv=rxv_spin_render(cond->pos,extra))!=APR_SUCCESS)
              return rv;
          } else if(cond->neg){
            if((rv=rxv_spin_render(cond->neg,extra))!=APR_SUCCESS)
              return rv;
          }
        } else if(cond->neg){
          if((rv=rxv_spin_render(cond->neg,extra))!=APR_SUCCESS)
            return rv;
        }
        break;
    }
  }

  return APR_SUCCESS;
}
